-- Author:          Ernest Jamro
-- OPB bus connected to the external SRAM memory (SRAM is 16 bit wide)

--//////////////////////////////////////////////////////////////////////
--//// Copyright (C) 2003 Authors                                   //// 
--////                                                              ////
--//// This source file may be used and distributed without         ////
--/// restriction provided that this copyright statement is not     ////
--//// removed from the file and that any derivative work contains  ////
--//// the original copyright notice and the associated disclaimer. ////
--////                                                              ////
--//// This source file is free software; you can redistribute it   ////
--//// and/or modify it under the terms of the GNU Lesser General   ////
--//// Public License as published by the Free Software Foundation; ////
--//// either version 2.1 of the License, or (at your option) any   ////
--//// later version.                                               ////
--////                                                              ////
--//// This source is distributed in the hope that it will be       ////
--//// useful, but WITHOUT ANY WARRANTY; without even the implied   ////
--//// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      ////
--//// PURPOSE. See the GNU Lesser General Public License for more  ////
--//// details.                                                     ////
--//////////////////////////////////////////////////////////////////////

library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_unsigned.all;

library UNISIM;
use unisim.all; -- for xilinx component signals


entity OPB_SRAM is
  generic (
    C_OPB_AWIDTH : integer                   := 24;
    C_OPB_DWIDTH : integer                   := 32;
    C_BASEADDR   : std_logic_vector(0 to 31) := X"0000_0000";
    C_HIGHADDR   : std_logic_vector(0 to 31) := X"0000_01FF";
	  C_SPEEDRATE  : integer := 0;
	  C_REAL_DWIDTH : integer:= 32; -- real width of the data -bus  the rst LSBs are always 0s
	  C_SRAM_DWIDTH: integer:= 16; -- external memory data width
	  C_SRAM_AWIDTH: integer:= 18; -- external memory address width
	  C_USE_SeqAddr: integer:= 1 -- use OPB_SeqAddr to speed up reading
    );
  port (
    -- Global signals
    OPB_Clk : in std_logic;
    OPB_Rst : in std_logic; -- not used in the project
    -- OPB signals
    OPB_ABus    : in std_logic_vector(0 to C_OPB_AWIDTH-1);
    OPB_BE      : in std_logic_vector(0 to C_OPB_DWIDTH/8-1);
    OPB_RNW     : in std_logic;
    OPB_select  : in std_logic;
    OPB_seqAddr : in std_logic;
    OPB_DBus    : in std_logic_vector(0 to C_OPB_DWIDTH-1);
    Sl_DBus    : out std_logic_vector(0 to C_OPB_DWIDTH-1);
    Sl_errAck  : out std_logic;
    Sl_retry   : out std_logic;
    Sl_toutSup : out std_logic;
    Sl_xferAck : out std_logic;
	  -- external SRAM signals
    pb_a: out STD_LOGIC_VECTOR (0 to c_sram_awidth-1); -- address bus
    pb_d_i: in STD_LOGIC_VECTOR (0 to c_sram_dwidth-1); -- data bus
	  pb_d_o: out STD_LOGIC_VECTOR (0 to c_sram_dwidth-1); -- data bus
	  pb_d_t: out STD_LOGIC; -- tri-state datap_o
    pb_oen: out std_logic; -- output enable (active low)
    pb_wen: out std_logic; -- write enable
    ram_cen: out std_logic; -- chip enable (always zero)
    pb_ubn: out std_logic; -- upper byte enable active low
    pb_lbn: out std_logic -- lower byte enable 
    );
end entity OPB_SRAM;

architecture opb_sram_arch of opb_sram is
-------------------------------------------------------------------------
-- functions and components declarations
function log2(arg: integer) return integer is
    variable ret: integer:= -1;
  begin
	if arg= 0 then
		ret:= -1;
		return ret;
	end if; 
	for i in 0 to 32 loop
	  if 2**i = arg then
		  ret:= i;
		  return i;
	  end if;
	end loop;
	assert false -- always assert when loop does not finish the function
	report "opb_sram. Log2() argument is not a power of 2"
	severity failure;
	return -1;
end log2;

  component ROC -- reset on Configuration
   port (O : out std_logic);
  end component;
component sram2opb
  generic (
  	C_AWIDTH: integer:= 18; -- number of address bits of the OPB bus connected to the SRAM= C_SRAM_AWIDTH - log2(C_DWIDTH/C_SRAM_DWIDTH)
    C_DWIDTH : integer:= 32; -- real width of the OPB data -bus  the rest MSBs are always 0s
    C_SRAM_DWIDTH: integer:= 16; -- external memory data width
	  C_SRAM_AWIDTH: integer:= 19; -- external memory address width
	  C_USE_SeqAddr: integer:= 0); -- use OPB_SeqAddr to speed up reading
  port (
  -- Global signals
    arst: in std_logic;
    clk : in std_logic;
    ABus    : in std_logic_vector(0 to c_awidth-1);
    DBus    : in std_logic_vector(0 to C_DWIDTH-1);
    Sl_DBus    : out std_logic_vector(0 to C_DWIDTH-1);
	  -- internal signals diffrent from OPB
	be: in std_logic_vector(0 to c_dwidth/8 -1); -- byte enable
	we: in std_logic; -- write enable
	rd: in std_logic; -- read command and device is selected
	xferAck: buffer std_logic; -- TRANSFER acknowledge
	ce_adr: in std_logic; -- address register clock enable
	seqAddr: in std_logic; -- increament address by 1 - valid only when C_USE_SeqAddr= 1
	-- external SRAM signals
  adr: out STD_LOGIC_VECTOR (0 to c_sram_awidth-1); -- address bus
  data_i: in STD_LOGIC_VECTOR (0 to c_sram_dwidth-1); -- data bus
	data_o: out STD_LOGIC_VECTOR (0 to c_sram_dwidth-1); -- data bus
	data_t: out STD_LOGIC; -- tri-state datap_o
  oen: out std_logic; -- output enable (active low)
  wen: out std_logic; -- write enable
  cen: out std_logic; -- chip enable (always zero)	
  ubn: out std_logic; 
  lbn: out std_logic
    );
end component;
component OPB_slave1
  generic (
    C_OPB_AWIDTH : integer                   := 32;
    C_OPB_DWIDTH : integer                   := 32;
    C_BASEADDR   : std_logic_vector          := X"0000_0100";
    C_HIGHADDR   : std_logic_vector          := X"0000_01FF";
	  C_SPEEDRATE : integer := 0; -- the lower the speed the quicker (lower number of clock cycles - latency) 
	   -- and the lower occupied area however the lower clock frequency is required. 
     -- Therefore to increase the clock frequency this parameter should be increased
	  -- user interface data and address width
    USE_SeqAddr : integer:= 1; -- use seq_addr (=0 saves area but reduces the transfer speed for seq_addr)
	  DWIDTH : integer:= 8; -- real width of the data - the rest LSBs are always 0s
	  AWIDTH : integer:= 8 -- user address width
    );
  port ( clk : in std_logic; -- clock signal
    arst: in std_logic; -- asynchronous reset (needed only for simulation)
    -- OPB signals
    OPB_ABus    : in std_logic_vector(0 to C_OPB_AWIDTH-1);
    OPB_BE      : in std_logic_vector(0 to C_OPB_DWIDTH/8-1);
    OPB_RNW     : in std_logic;
    OPB_select  : in std_logic;
    OPB_seqAddr : in std_logic;
    OPB_DBus    : in std_logic_vector(0 to C_OPB_DWIDTH-1);
    Sl_DBus    : out std_logic_vector(0 to C_OPB_DWIDTH-1);
    Sl_errAck  : out std_logic;
    Sl_retry   : out std_logic;
    Sl_toutSup : out std_logic;
    Sl_xferAck : out std_logic;
	  -- user signals
	  sel: out std_logic; -- select - active when OPB_select and proper address
	  ack: in std_logic; -- acknowlenge - slave is ready for data transfer
	  rnw: out std_logic; -- read not write - same assert OPB_read not write
	  adr: out std_logic_vector(0 to awidth-1); -- address bus
	  drd: in std_logic_vector(0 to dwidth-1); -- data read by opb master (send by the uer device)
	  dwr: out std_logic_vector(0 to dwidth-1); -- data written by the master (recieved by the user device)
	  be: out std_logic_vector(0 to dwidth/8 -1); -- byte enable 
	  seqAddr: out std_logic -- same as OPB_seqAddr
    );
   end component;

-----------------------------------------------------------------------------------------------
  -- constants and signals declarations
  constant adr_space: integer:= conv_integer(c_highaddr - c_baseaddr + 1);
  constant awidth: integer:= log2(adr_space); -- address width when address spac is considered
  constant ignore_adr: integer:= log2(c_real_dwidth/8); -- how many LSBs of the address will be ignored (because of the data size)
  constant real_awidth: integer:= awidth - ignore_adr; -- real address lines that must be considered
  constant ABus_width: integer:= c_sram_awidth - log2(c_real_dwidth/c_sram_dwidth); -- address width that is send to sram2opb block  

  signal adrs: std_logic_vector(0 to awidth-1); -- address bus that is directly copied from OPB
  signal ABus: std_logic_vector(0 to abus_width-1); -- sram  module address bus
  signal BE, bes: std_logic_vector(0 to c_real_dwidth/8 -1); -- Byte enable (excluding not used bits)
  
  signal rd, we: std_logic; -- write enable and read signal
  
  signal ce_adr: std_logic; -- clock enable for address
  signal arst: std_logic; -- asynchronous global set/ reset
  signal ack, sel, rnw, seqAddr: std_logic; -- internal opb_slave signals
  signal write_single_byte: std_logic; -- read cycle before write
  signal drd, dwr, DBus: std_logic_vector(0 to c_real_dwidth-1); -- internal signals drd, dwr - form opb_slave, DBus- dwr for rd_before_wr=1
begin  
 sl: OPB_slave1
  generic map (C_OPB_AWIDTH=> C_OPB_AWIDTH, C_OPB_DWIDTH=> C_OPB_DWIDTH,
    C_BASEADDR=> C_BASEADDR, C_HIGHADDR=> C_HIGHADDR, C_SPEEDRATE=> C_SPEEDRATE,
	DWIDTH=> C_REAL_DWIDTH, AWIDTH=> awidth)
  port map(clk=> OPB_clk, arst=> arst,
    -- OPB signals
    OPB_ABus=> OPB_ABus, OPB_BE=> OPB_BE, OPB_RNW=> OPB_RNW,
    OPB_select=> OPB_select, OPB_seqAddr=> OPB_seqAddr,
    OPB_DBus=> OPB_DBus, Sl_DBus=> Sl_DBus, Sl_errAck=> Sl_errAck,
    Sl_retry=> Sl_retry, Sl_toutSup=> Sl_toutSup, Sl_xferAck=> Sl_xferAck,
	-- user signals
	sel=> sel, ack=> ack, rnw=> rnw, adr=> adrs,
	drd=> drd, dwr=> dwr, be=> bes, seqAddr=> seqAddr);
	
  gadr: if real_awidth<ABus_width generate -- FILL not used address lines with zeros (not all memory address lines are used
	  ABus(0 to ABus_width-real_awidth-1)<= (others=>'0'); -- set not used MSBs address lines lines to zero
  end generate; 
  ABus(ABus_width-real_awidth to ABus_width-1)<= -- copy address bits from OPB_ABus
     adrs(0 to awidth-ignore_adr-1);		
 
  s2o: sram2opb 
    generic map(C_AWIDTH=> abus_width, C_DWIDTH=> c_real_dwidth,
	  C_SRAM_DWIDTH=> c_sram_dwidth, C_SRAM_AWIDTH=> c_sram_awidth,
	  C_USE_SeqAddr=> c_use_seqAddr)
    port map(arst=> arst, clk=> OPB_clk, 
       ABus=> ABus, DBus=> DBus, Sl_DBus=> drd,
	   be=> BE, we=> we, rd=> rd, xferAck=> Ack, 
	   ce_adr=> ce_adr, seqAddr=> OPB_SeqAddr,
	  -- external SRAM signals
      adr=> pb_a, data_i=> pb_d_i, data_o=> pb_d_o, data_t=> pb_d_t,
      oen=> pb_oen, wen=> pb_wen, cen=> ram_cen, lbn=> pb_lbn, ubn => pb_ubn);

 ce_adr<= sel;
	  
	  
	 rd<= sel and RNW;
	 we<= sel and not RNW;
   BE<= bes;
	 DBus<= dwr;

-- crd1: if C_RD_BEFORE_WR>0 and c_sram_dwidth= 16 generate
--	g16: if c_real_dwidth = 16 generate
--		 write_single_byte<= (BES(0) xor BES(1)) 
--		   and sel and not RNW;
--	    BE<= "11";
--        -- input multiplexer separate for each data
--	end generate;
--	 
--	g32: if c_real_dwidth= 32 generate
--		write_single_byte <= ((BES(0) xor BES(1) ) or (BES(2) xor BES(3)) )
--		   and sel and not RNW;
--        BE(0)<= BES(0) or BES(1);
--	    BE(1)<= BE(0);
--	    BE(2)<= BES(2) or BES(3); 
--	    BE(3)<= BE(2);
--    end generate;
--
--    gdin: for i in 0 to c_real_dwidth/8-1 generate
--	  DBus(8*i to 8*i+7)<= dwr(8*i to 8*i+7)
--            when BES(i)= '1' else
--          drd(8*i to 8*i+7); -- valid for read before write logic
--    end generate;
--	
--	rd<= write_single_byte or (sel and RNW);
--	we<= Ack when write_single_byte='1' else
--		sel and not RNW;
--	
-- end generate;
 
 -- global set/reset logic - this component draws rst during chip configuration only (no external reset is used)
 global_reset: ROC
	port map (O=> arst);

 assert c_real_dwidth<=c_opb_dwidth
   report "opb_sram error. c_real_dwidth must be greater than c_opb_dwidth"
   severity error;
assert c_real_dwidth>=c_sram_dwidth
   report "opb_sram error. c_real_dwidth must be greater or eq. to the c_sram_dwidth"
   severity error;
 assert adr_space*c_real_dwidth/c_opb_dwidth <= (2**c_sram_awidth) * c_sram_dwidth/8
   report "opb_sram error. The memory address space is too large in comparison to the actual memory size"
   severity error;
   
end opb_sram_arch;